#include <iostream.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>
#include <signal.h>

#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/soundcard.h>

#include <kapp.h>
#include <qpalette.h>
#include <qcolor.h>
#include <qlayout.h>
#include <qlabel.h>

#include "kntv.h"
#include "videodev.h"

// Capture size (scales to fit window - don't mess with this)
#define MY_WIDTH 480
#define MY_HEIGHT 360

typedef struct tag_vimage
{
	struct video_buffer     vidbuf;
	char                    *data;
} VIMAGE;

// global variables
int vol = 75;			// current sound volume
int cur_width = MY_WIDTH;	// current video width
int cur_height = MY_HEIGHT;	// current video height
int mixer_fd = -1;		// the mixer file descriptor
int vid_fd = -1;		// the video input file descriptor
int source;			// mixer input source
unsigned supported;		// soundcard supported features bitfield

Kntv::Kntv()
{
	char buf[8];
	
	this->resize(cur_width,cur_height);
	
	// the video window, filled with our chroma color (0xff00ff)
	QLabel *box = new QLabel(this,"box",0U);
	box->setPalette(QPalette(QColor(255,0,255)));
	box->resize(cur_width,cur_height);
	
	// a vertical layout to hold our window and buttons
	QVBoxLayout *vbox = new QVBoxLayout(this);
	vbox->addWidget(box,0,0);
	
	// the slider
	QHBoxLayout *hbox = new QHBoxLayout();
	vbox->addLayout(hbox,0);
	QLabel *lb = new QLabel("Volume: ",this,0);
	lb->setMinimumSize(lb->sizeHint());
	lb->setMaximumHeight(lb->height());
	volSlider = new QSlider(0,100,5,75,QSlider::Horizontal,this,"vslider");
	volSlider->setTickmarks(QSlider::Below);
	volSlider->setMaximumHeight(lb->height());
	hbox->addWidget(lb,0,0);
	hbox->addWidget(volSlider,1,0);
	
	connect(volSlider, SIGNAL(valueChanged(int)), SLOT(volChangeCallback(int)));
}

Kntv::~Kntv()
{
}

void Kntv::MoveVidcapWindow(int x, int y)
{
	struct video_window vw = 
		{x,y,cur_width,cur_height,0xff00ff,0,0};
	ioctl(vid_fd, VIDIOC_S_WIN, &vw);
}

void Kntv::setSource(int id)
{
	source = 1 << id;
	if (ioctl (mixer_fd, SOUND_MIXER_WRITE_RECSRC, &source) < 0)
		perror ("Mixer.set_source()");
}

void Kntv::setVolume(int mixer_id, int volume)
{
	if (ioctl (mixer_fd, MIXER_WRITE (mixer_id), &volume) < 0)
		perror ("Mixer.set_mixer()");
}

void Kntv::volChangeCallback(int v)
{
	vol = v;
	setVolume(SOUND_MIXER_IMIX, (vol<<8)+vol);
}

// Set mixer input source to 0, zero rec mon volume and close descriptors
void shutdown(int sig)
{
	int parm = 0;
	ioctl(mixer_fd, MIXER_WRITE (SOUND_MIXER_IMIX), &parm);
	ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &parm);
	close(mixer_fd);
	close(vid_fd);
	exit(0);
}

// Clean up and exit
void Kntv::closeEvent(QCloseEvent *e)
{
	shutdown(0);
}

// Handle window movement
void Kntv::moveEvent(QMoveEvent *e)
{
	QPoint pos = e->pos();
	this->MoveVidcapWindow(pos.x(), pos.y());
	this->repaint();
}

// Handle window resizing
void Kntv::resizeEvent(QResizeEvent *e)
{
	QPoint pos = this->pos();
	QSize sz = e->size();
	cur_width = sz.width()+10;
	cur_height = sz.height();
	this->MoveVidcapWindow(pos.x(),pos.y());
	this->repaint();
}

int main(int argc, char **argv)
{
	int err,i;
	struct video_format fmt;
	struct video_parm parm;
	struct video_standard std;
	VIMAGE vimage;
	struct video_framebuffer vf;

	// Die gracefully on ctrl-c and term signals
	signal(SIGINT, &shutdown);
	signal(SIGTERM, &shutdown);

	// Set up the KDE windows and app context
	KApplication *app = new KApplication(argc, argv);

	// Open video device
	vid_fd = open("/dev/video0", O_RDONLY);
	if(vid_fd < 0) {
		perror("Vidcap.open");
		return -1;
	}
	
	// Open mixer and set volume
	mixer_fd = open("/dev/mixer", O_RDWR);
	if(mixer_fd < 0) {
		perror("Mixer.open");
		return -1;
	}

	if(ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &supported) < 0)
		supported = 0xffff;

//->    Set video params 0=MXC, 1=COMPOSITE, 2=SVIDEO
	ioctl(vid_fd, VIDIOC_G_PARM, &parm);
        if (err) {
		printf("VIDIOC_G_PARM: error %d.\n",err);
	}
	parm.input = 0;
	err = ioctl(vid_fd, VIDIOC_S_PARM, &parm);
        if (err) {
		printf("VIDIOC_S_PARM: error %d.\n",err);
	}
	
#if 0
// wms: this call is not supported in the driver yet - skip it
//->    Set video standard VIDEO_STD_PAL or VIDEO_STD_NTSC
	err = ioctl(vid_fd, VIDIOC_S_STD, &std);
        if (err) {
		printf("VIDIOC_S_SID: error %d.\n",err);
//                return -1;
	}
#endif

//->    Image width, height, pixel format
	ioctl(vid_fd, VIDIOC_G_FMT, &fmt);
        fmt.width = MY_WIDTH;
        fmt.height = MY_HEIGHT;
        fmt.pixelformat = PIX_FMT_RGB565;

	/* wms: try to iterate through some picture sizes... */
	err = 1;
	while(fmt.width>0 && err)
	{
	    printf("Trying for picture size %dx%d... ",fmt.width, fmt.height);
	    err = ioctl (vid_fd, VIDIOC_S_FMT, &fmt);
	    if (err)
	    {
		fmt.width -= 40;
		fmt.height -= 30;
		printf(" VIDIOC_S_FMT err %d.\n",err);
	    }
	}
    
	if (err)
	{
	    printf("Could not allocate video memory for vidcap. Aborting...\n"); 
	    return(1);
	}	    
	else
	    printf(" OK.\n");

	cur_width = fmt.width;	// current video width
	cur_height = fmt.height;// current video height

	ioctl(vid_fd, VIDIOC_G_FBUF, &vf);
	vf.flags = FBUF_FLAG_CHROMAKEY | FBUF_FLAG_OVERLAY;
	err = ioctl(vid_fd, VIDIOC_S_FBUF, &vf);
        if (err) {
		printf("VIDIOC_S_FBUF: error %d.\n",err);
	}
	err = ioctl(vid_fd, VIDIOC_PREVIEW, 1);
        if (err) {
		printf("VIDIOC_PREVIEW: error %d.\n",err);
	}

	vimage.data = (char *)malloc(fmt.sizeimage);
	if(vimage.data == 0) {
		cout << "Unable to allocate memory for video buffer." << endl;
		return -1;
	}

	Kntv *tv = new Kntv();
	app->setMainWidget(tv);
	tv->show();

	tv->setSource(6); // turn on record monitor on line in
	tv->setVolume(SOUND_MIXER_IMIX, (vol<<8)+vol); // set source volume
	
	return app->exec();
}

#include "kntv.moc"
